// *****************************************************************************
//  Title.............. :  The  Serial Port Synaser Daemon
//
//  Modulname ......... :  serialPortDaemon.lpr (project file)
//  Type .............. :  Unit
//  Author ............ :  Udo Schmal  - J.P
//  Development Status  :  01.12.2016
//  Operating System .. :  Win32/Win64
//  IDE ............... :  Lazarus
// *****************************************************************************
program serialPortdaemon;

{$mode objfpc}{$H+}

uses
  HeapTrc,
{$IFDEF WINDOWS} Windows,  {$ENDIF}
{$IFDEF UNIX}{$IFDEF UseCThreads}
  CThreads,
{$ENDIF} Cmem,{$ENDIF}
  Classes, SysUtils, EventLog, DaemonApp,synaser,fpTimer;

type
  TSpDaemon  = class;
  TTheThread = class(TThread)
     procedure Execute; override;
     destructor Destroy; override;
  end;
  { TComPortReadThread }
  TComPortReadThread=class(TThread)
  public
    MustDie: boolean;
    Owner: TSpDaemon;
  protected
    procedure CallEvent;
    procedure Execute; override;
  published
    property Terminated;
  end;

  TSpDaemon = class(TCustomDaemon)
  private
    FThread: TTheThread;
    FReadThread: TComPortReadThread;
  public
    CurPos : integer;
    FTempStr: String;
    FRcvLineCRLF : Boolean;
    FSynSer: TBlockSerial;
    OnRxData: TNotifyEvent;
    function Install: boolean; override;
    function UnInstall: boolean; override;
    function Start: boolean; override;
    function Stop: boolean; override;
    function Pause: boolean; override;
    function Continue: boolean; override;
    function Execute: boolean; override;
    function ShutDown: boolean; override;
    function DataAvailable: boolean;
    function ReadData: string;
  end;

  TSpDaemonMapper = class(TCustomDaemonMapper)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ToDoOnInstall(Sender: TObject);
    procedure ToDoOnRun(Sender: TObject);
    procedure ToDoOnUninstall(Sender: TObject);
    procedure ToDoOnDestroy(Sender: TObject);
  end;

function BoolToStr(AVal: Boolean): String;
begin
  if AVal = True then result := 'true' else result := 'false';
end;

procedure TTheThread.Execute;
var i: integer;
begin
  i := 0;
  Application.Log(etDebug, 'Thread.Execute');
  try
    repeat
      Sleep(5000); //milliseconds
      inc(i);
      Application.Log(etDebug, 'Thread.Loop ' + Format('Tick :%d', [i]));
    until Terminated;
  finally
    Application.Log(etDebug, 'Thread.LoopStopped');
  end;
end;

destructor TTheThread.Destroy;
begin
  Application.Log(etDebug, 'Thread.Destroy');
  inherited Destroy;
end;
{ TComPortReadThread }

procedure TComPortReadThread.CallEvent;
begin
  if Assigned(Owner.OnRxData) then begin
    Owner.OnRxData(Owner);
  end;
end;

procedure TComPortReadThread.Execute;
var Str : string;
begin
  Application.Log(etDebug, 'TComPortReadThread.Execute');
  try
    while not MustDie do begin
      if Owner.FSynSer.CanReadEx(100) then
          begin
            Str :=  Owner.ReadData;
            if Str <> '' Then Application.Log(etDebug, Str);
//        Synchronize(@CallEvent);
          end;

    end;
  finally
    Application.Log(etDebug, 'TComPortReadThread.Terminate');
    Terminate;
  end;

end;

{$REGION ' - Daemon - '}
function TSpDaemon.Install: boolean;
begin
  result := inherited Install;
  Application.Log(etDebug, 'Daemon.installed: ' + BoolToStr(result));
end;

function TSpDaemon.UnInstall: boolean;
begin
  result := inherited UnInstall;
  Application.Log(etDebug, 'Daemon.Uninstall: ' + BoolToStr(result));
end;
function TSpDaemon.Start: boolean;
begin
  result := inherited Start;
  Application.Log(etDebug, 'Daemon.Start: ' + BoolToStr(result));
//  FThread := TTheThread.Create(true);
//  FThread.FreeOnTerminate := true;
//  FThread.Resume;
  FSynSer := TBlockSerial.Create;
  FSynSer.Connect('COM3');
  if FSynSer.Handle=INVALID_HANDLE_VALUE then
    begin
      Application.Log(etDebug,'Could not open device '+ FSynSer.Device);
    end
  else
    begin
     FRcvLineCRLF := True;
     FSynSer.Config(9600, 8, 'N', SB1, False, False);
     FReadThread := TComPortReadThread.Create(true);
     FReadThread.FreeOnTerminate := true;
     FReadThread.Owner := Self;
     FReadThread.MustDie := false;
     FReadThread.Start;
    end;
end;


function TSpDaemon.Stop: boolean;
begin
  result := inherited Stop;
  Application.Log(etDebug, 'Daemon.Stop: ' + BoolToStr(result));
//  FThread.Terminate;
//  FThread.WaitFor;
//  FThread := nil;
   // flush device
  if FSynSer.Handle<>INVALID_HANDLE_VALUE then begin
    FSynSer.Flush;
    FSynSer.Purge;
  end;

  // stop capture thread
  if FReadThread<>nil then begin
    FReadThread.FreeOnTerminate:=false;
    FReadThread.MustDie:= true;
    while not FReadThread.Terminated do begin
      Sleep(1000);
      //Application.ProcessMessages;
    end;
    FReadThread.Free;
    FReadThread:=nil;
  end;

  // close device
  if FSynSer.Handle<>INVALID_HANDLE_VALUE then begin
    FSynSer.Flush;
    FSynSer.CloseSocket;
  end;
end;

function TSpDaemon.Pause: boolean;
begin
  result := inherited Pause;
  Application.Log(etDebug, 'Daemon.Pause: ' + BoolToStr(result));
//  FThread.Suspend;
  FReadThread.Suspend;
end;

function TSpDaemon.Continue: boolean;
begin
  result := inherited Continue;
  Application.Log(etDebug, 'Daemon.Continue: ' + BoolToStr(result));
//  FThread.Resume;
  FReadThread.Resume;
end;

function TSpDaemon.Execute: boolean;
begin
  result := inherited Execute;
  Application.Log(etDebug, 'Daemon.Execute: ' + BoolToStr(result));
end;

function TSpDaemon.ShutDown: boolean;
begin
  result := inherited ShutDown;
  Application.Log(etDebug, 'Daemon.ShutDown: ' + BoolToStr(result));
end;
function TSpDaemon.DataAvailable: boolean;
begin
  if FSynSer.Handle=INVALID_HANDLE_VALUE then begin
    result:=false;
    exit;
  end;
  result:=FSynSer.CanReadEx(0);
end;
function TSpDaemon.ReadData: string;
begin
  result:='';
  if FSynSer.Handle=INVALID_HANDLE_VALUE then
    Application.Log(etDebug,'can not read from a closed port.');
  if FRcvLineCRLF then
  result:=FSynSer.RecvString(0)
  else
  result:=FSynSer.RecvPacket(0);
end;
{$ENDREGION}

{$REGION ' - DaemonMapper - '}
constructor TSpDaemonMapper.Create(AOwner: TComponent);
begin
  Application.Log(etDebug, 'SpDaemonMapper.Create');
  inherited Create(AOwner);
  with DaemonDefs.Add as TDaemonDef do
  begin
    DaemonClassName := 'TSpDaemon';
    Name := 'spDaemon';
    Description := 'Synaser Read Serial Port Demo';
    DisplayName := 'Serial Port Daemon';
    RunArguments := '--run';
    Options := [doAllowStop,doAllowPause];
    Enabled := true;
    with WinBindings do
    begin
      StartType := stBoot;
      WaitHint := 0;
      IDTag := 0;
      ServiceType := stWin32;
      ErrorSeverity := esNormal;//esIgnore;
    end;
//    OnCreateInstance := ?;
    LogStatusReport := false;
  end;
  OnInstall := @Self.ToDoOnInstall;
  OnRun := @Self.ToDoOnRun;
  OnUnInstall := @Self.ToDoOnUninstall;
  OnDestroy := @Self.ToDoOnDestroy;
  Application.Log(etDebug, 'SpDaemonMapper.Created');
end;

procedure TSpDaemonMapper.ToDoOnInstall(Sender: TObject);
begin
  Application.Log(etDebug, 'SpDaemonMapper.Install');
end;

procedure TSpDaemonMapper.ToDoOnRun(Sender: TObject);
begin
  Application.Log(etDebug, 'SpDaemonMapper.Run');
end;

procedure TSpDaemonMapper.ToDoOnUnInstall(Sender: TObject);
begin
  Application.Log(etDebug, 'SpDaemonMapper.Uninstall');
end;

procedure TSpDaemonMapper.ToDoOnDestroy(Sender: TObject);
begin
  //doesn't comes here
  Application.Log(etDebug, 'SpDaemonMapper.Destroy');
end;
{$ENDREGION}

{$R fclel.res}

begin
  RegisterDaemonClass(TSpDaemon);
  RegisterDaemonMapper(TSpDaemonMapper);
  RegisterDaemonApplicationClass(TCustomDaemonApplication);
  heaptrc.SetHeapTraceOutput(ChangeFileExt(ParamStr(0), '.heap'));
  with Application do
  begin
    Title := 'Serial Port Daemon Application';
//    EventLog.LogType := ltSystem;
    EventLog.LogType := ltFile;
    EventLog.DefaultEventType := etDebug;
    EventLog.AppendContent := true;
    EventLog.FileName := ChangeFileExt(ParamStr(0), '.log');
    Initialize;
    Run;
  end;
end.
